﻿using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace Proxmox.SpiceLauncher;

static class DPApi
{
    private static int EntropySize = 16;
    public static byte[] CreateRandomEntropy()
    {
        // Create a byte array to hold the random value.
        var entropy = new byte[EntropySize];

        // Create a new instance of the RNGCryptoServiceProvider.
        // Fill the array with a random value.
        RandomNumberGenerator.Create().GetBytes(entropy);

        // Return the array.
        return entropy;
    }

    public static byte[] EncryptString(string s)
    {
        var data = System.Text.Encoding.UTF8.GetBytes(s);
        var ms = new MemoryStream();
        var entropy = CreateRandomEntropy();
        ms.Write(entropy);
        EncryptDataToStream(data, entropy, DataProtectionScope.CurrentUser, ms);
        return ms.ToArray();
    }

    public static string EncryptStringToBase64(string s)
    {
        var data = EncryptString(s);
        return Convert.ToBase64String(data);
    }

    public static string DecryptString(byte[] data)
    {
        var ms = new MemoryStream(data);
        var entropy = new byte[EntropySize];
        var entropySize = ms.Read(entropy);
        if (entropySize != EntropySize)
            throw new IOException("Entropy size was not correct.");
        var dec = DecryptDataFromStream(entropy, DataProtectionScope.CurrentUser, ms, data.Length - entropySize);
        return Encoding.UTF8.GetString(dec);
    }

    public static string DecryptStringFromBase64(string s)
    {
        var data = Convert.FromBase64String(s);
        return DecryptString(data);
    }

    public static int EncryptDataToStream(byte[] buffer, byte[] entropy, DataProtectionScope scope, Stream s)
    {
        if (buffer == null)
            throw new ArgumentNullException(nameof(buffer));
        if (buffer.Length <= 0)
            throw new ArgumentException("The buffer length was 0.", nameof(buffer));
        if (entropy == null)
            throw new ArgumentNullException(nameof(entropy));
        if (entropy.Length <= 0)
            throw new ArgumentException("The entropy length was 0.", nameof(entropy));
        if (s == null)
            throw new ArgumentNullException(nameof(s));

        var length = 0;

        // Encrypt the data and store the result in a new byte array. The original data remains unchanged.
        var encryptedData = ProtectedData.Protect(buffer, entropy, scope);

        // Write the encrypted data to a stream.
        if (s.CanWrite && encryptedData != null)
        {
            s.Write(encryptedData, 0, encryptedData.Length);

            length = encryptedData.Length;
        }

        // Return the length that was written to the stream.
        return length;
    }

    public static byte[] DecryptDataFromStream(byte[] entropy, DataProtectionScope scope, Stream s, int length)
    {
        if (s == null)
            throw new ArgumentNullException(nameof(s));
        if (length <= 0)
            throw new ArgumentException("The given length was 0.", nameof(length));
        if (entropy == null)
            throw new ArgumentNullException(nameof(entropy));
        if (entropy.Length <= 0)
            throw new ArgumentException("The entropy length was 0.", nameof(entropy));

        var inBuffer = new byte[length];
        byte[] outBuffer;

        // Read the encrypted data from a stream.
        if (s.CanRead)
        {
            s.Read(inBuffer, 0, length);

            outBuffer = ProtectedData.Unprotect(inBuffer, entropy, scope);
        }
        else
        {
            throw new IOException("Could not read the stream.");
        }

        // Return the decrypted data
        return outBuffer;
    }
}